# -*- coding: utf-8 -*-

# Automatically generated - don't edit.
# Use `python setup.py build_ui` to update it.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_TagsCompatibilityOptionsPage(object):
    def setupUi(self, TagsCompatibilityOptionsPage):
        TagsCompatibilityOptionsPage.setObjectName("TagsCompatibilityOptionsPage")
        TagsCompatibilityOptionsPage.resize(539, 705)
        self.vboxlayout = QtWidgets.QVBoxLayout(TagsCompatibilityOptionsPage)
        self.vboxlayout.setObjectName("vboxlayout")
        self.wave_files = QtWidgets.QGroupBox(TagsCompatibilityOptionsPage)
        self.wave_files.setObjectName("wave_files")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.wave_files)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.label = QtWidgets.QLabel(self.wave_files)
        self.label.setWordWrap(True)
        self.label.setObjectName("label")
        self.verticalLayout_3.addWidget(self.label)
        spacerItem = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.verticalLayout_3.addItem(spacerItem)
        self.write_wave_riff_info = QtWidgets.QCheckBox(self.wave_files)
        self.write_wave_riff_info.setObjectName("write_wave_riff_info")
        self.verticalLayout_3.addWidget(self.write_wave_riff_info)
        self.remove_wave_riff_info = QtWidgets.QCheckBox(self.wave_files)
        self.remove_wave_riff_info.setObjectName("remove_wave_riff_info")
        self.verticalLayout_3.addWidget(self.remove_wave_riff_info)
        self.wave_riff_info_encoding = QtWidgets.QGroupBox(self.wave_files)
        self.wave_riff_info_encoding.setObjectName("wave_riff_info_encoding")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout(self.wave_riff_info_encoding)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.wave_riff_info_enc_cp1252 = QtWidgets.QRadioButton(self.wave_riff_info_encoding)
        self.wave_riff_info_enc_cp1252.setChecked(True)
        self.wave_riff_info_enc_cp1252.setObjectName("wave_riff_info_enc_cp1252")
        self.horizontalLayout_3.addWidget(self.wave_riff_info_enc_cp1252)
        self.wave_riff_info_enc_utf8 = QtWidgets.QRadioButton(self.wave_riff_info_encoding)
        self.wave_riff_info_enc_utf8.setObjectName("wave_riff_info_enc_utf8")
        self.horizontalLayout_3.addWidget(self.wave_riff_info_enc_utf8)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem1)
        self.verticalLayout_3.addWidget(self.wave_riff_info_encoding)
        self.vboxlayout.addWidget(self.wave_files)
        spacerItem2 = QtWidgets.QSpacerItem(274, 41, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.vboxlayout.addItem(spacerItem2)

        self.retranslateUi(TagsCompatibilityOptionsPage)
        self.write_wave_riff_info.toggled['bool'].connect(self.remove_wave_riff_info.setDisabled) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(TagsCompatibilityOptionsPage)

    def retranslateUi(self, TagsCompatibilityOptionsPage):
        _translate = QtCore.QCoreApplication.translate
        self.wave_files.setTitle(_("WAVE files"))
        self.label.setText(_("Picard will tag WAVE files using ID3v2 tags. This is not supported by all software. For compatibility with software which does not support ID3v2 tags in WAVE files additional RIFF INFO tags can be written to the files. RIFF INFO has only limited support for tags and character encodings."))
        self.write_wave_riff_info.setText(_("Also include RIFF INFO tags in the files"))
        self.remove_wave_riff_info.setText(_("Remove existing RIFF INFO tags from WAVE files"))
        self.wave_riff_info_encoding.setTitle(_("RIFF INFO text encoding"))
        self.wave_riff_info_enc_cp1252.setText(_("Windows-1252"))
        self.wave_riff_info_enc_utf8.setText(_("UTF-8"))



# fix-header: skip

# http://hetland.org/coding/python/levenshtein.py

# This is a straightforward implementation of a well-known algorithm, and thus
# probably shouldn't be covered by copyright to begin with. But in case it is,
# the author (Magnus Lie Hetland) has, to the extent possible under law,
# dedicated all copyright and related and neighboring rights to this software
# to the public domain worldwide, by distributing it under the CC0 license,
# version 1.0. This software is distributed without any warranty. For more
# information, see <http://creativecommons.org/publicdomain/zero/1.0>


def astrcmp_py(a, b):
    """Calculates the Levenshtein distance between a and b."""
    n, m = len(a), len(b)
    if n > m:
        # Make sure n <= m, to use O(min(n,m)) space
        a, b = b, a
        n, m = m, n

    if n == 0 or m == 0.0:
        return 0.0

    current = range(n+1)
    for i in range(1, m+1):
        previous, current = current, [i]+[0]*n
        for j in range(1, n+1):
            add, delete = previous[j]+1, current[j-1]+1
            change = previous[j-1]
            if a[j-1] != b[i-1]:
                change += 1
            current[j] = min(add, delete, change)

    return 1.0 - current[n] / max(m, n)


try:
    from picard.util._astrcmp import astrcmp as astrcmp_c
    astrcmp = astrcmp_c
    astrcmp_implementation = "C"
except ImportError:
    astrcmp = astrcmp_py
    astrcmp_implementation = "Python"



# -*- coding: utf-8 -*-

# Automatically generated - don't edit.
# Use `python setup.py build_ui` to update it.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_CaaOptions(object):
    def setupUi(self, CaaOptions):
        CaaOptions.setObjectName("CaaOptions")
        CaaOptions.resize(660, 194)
        self.verticalLayout = QtWidgets.QVBoxLayout(CaaOptions)
        self.verticalLayout.setObjectName("verticalLayout")
        self.select_caa_types_group = QtWidgets.QHBoxLayout()
        self.select_caa_types_group.setObjectName("select_caa_types_group")
        self.restrict_images_types = QtWidgets.QCheckBox(CaaOptions)
        self.restrict_images_types.setObjectName("restrict_images_types")
        self.select_caa_types_group.addWidget(self.restrict_images_types)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.select_caa_types_group.addItem(spacerItem)
        self.select_caa_types = QtWidgets.QPushButton(CaaOptions)
        self.select_caa_types.setEnabled(False)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(100)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.select_caa_types.sizePolicy().hasHeightForWidth())
        self.select_caa_types.setSizePolicy(sizePolicy)
        self.select_caa_types.setObjectName("select_caa_types")
        self.select_caa_types_group.addWidget(self.select_caa_types)
        self.verticalLayout.addLayout(self.select_caa_types_group)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.label = QtWidgets.QLabel(CaaOptions)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setObjectName("label")
        self.horizontalLayout.addWidget(self.label)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.cb_image_size = QtWidgets.QComboBox(CaaOptions)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.cb_image_size.sizePolicy().hasHeightForWidth())
        self.cb_image_size.setSizePolicy(sizePolicy)
        self.cb_image_size.setObjectName("cb_image_size")
        self.horizontalLayout.addWidget(self.cb_image_size)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.cb_approved_only = QtWidgets.QCheckBox(CaaOptions)
        self.cb_approved_only.setObjectName("cb_approved_only")
        self.verticalLayout.addWidget(self.cb_approved_only)
        spacerItem2 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem2)

        self.retranslateUi(CaaOptions)
        QtCore.QMetaObject.connectSlotsByName(CaaOptions)
        CaaOptions.setTabOrder(self.restrict_images_types, self.select_caa_types)
        CaaOptions.setTabOrder(self.select_caa_types, self.cb_image_size)
        CaaOptions.setTabOrder(self.cb_image_size, self.cb_approved_only)

    def retranslateUi(self, CaaOptions):
        _translate = QtCore.QCoreApplication.translate
        CaaOptions.setWindowTitle(_("Form"))
        self.restrict_images_types.setText(_("Download only cover art images matching selected types"))
        self.select_caa_types.setText(_("Select types…"))
        self.label.setText(_("Only use images of the following size:"))
        self.cb_approved_only.setText(_("Download only approved images"))



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2007 Lukáš Lalinský
# Copyright (C) 2014 Shadab Zafar
# Copyright (C) 2015-2021 Laurent Monin
# Copyright (C) 2019 Wieland Hoffmann
# Copyright (C) 2019-2020, 2022-2023 Philipp Wolfer
# Copyright (C) 2023 Bob Swift
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from functools import partial
import importlib
from importlib.abc import MetaPathFinder
import json
import os.path
import shutil
import sys
import tempfile
import zipfile
import zipimport

from PyQt5 import QtCore

from picard import log
from picard.const import (
    PLUGINS_API,
    USER_PLUGIN_DIR,
)
from picard.const.sys import IS_FROZEN
from picard.plugin import (
    _PLUGIN_MODULE_PREFIX,
    PluginData,
    PluginWrapper,
    _unregister_module_extensions,
)
import picard.plugins
from picard.version import (
    Version,
    VersionError,
)


_SUFFIXES = tuple(importlib.machinery.all_suffixes())
_PACKAGE_ENTRIES = ("__init__.py", "__init__.pyc", "__init__.pyo")
_PLUGIN_PACKAGE_SUFFIX = ".picard"
_PLUGIN_PACKAGE_SUFFIX_LEN = len(_PLUGIN_PACKAGE_SUFFIX)
_FILEEXTS = ('.py', '.pyc', '.pyo', '.zip')
_UPDATE_SUFFIX = '.update'
_UPDATE_SUFFIX_LEN = len(_UPDATE_SUFFIX)


_extension_points = []


def is_update(path):
    return path.endswith(_UPDATE_SUFFIX)


def strip_update_suffix(path):
    if not is_update(path):
        return path
    return path[:-_UPDATE_SUFFIX_LEN]


def is_zip(path):
    return path.endswith('.zip')


def strip_zip_suffix(path):
    if not is_zip(path):
        return path
    return path[:-4]


def is_package(path):
    return path.endswith(_PLUGIN_PACKAGE_SUFFIX)


def strip_package_suffix(path):
    if not is_package(path):
        return path
    return path[:-_PLUGIN_PACKAGE_SUFFIX_LEN]


def is_zipped_package(path):
    return path.endswith(_PLUGIN_PACKAGE_SUFFIX + '.zip')


def _plugin_name_from_path(path):
    path = os.path.normpath(path)
    if is_zip(path):
        name = os.path.basename(strip_zip_suffix(path))
        if is_package(name):
            return strip_package_suffix(name)
        else:
            return name
    elif os.path.isdir(path):
        for entry in _PACKAGE_ENTRIES:
            if os.path.isfile(os.path.join(path, entry)):
                return os.path.basename(path)
    else:
        file = os.path.basename(path)
        if file in _PACKAGE_ENTRIES:
            return None
        name, ext = os.path.splitext(file)
        if ext in _SUFFIXES:
            return name
        return None


def load_zip_manifest(archive_path):
    if is_zipped_package(archive_path):
        try:
            archive = zipfile.ZipFile(archive_path)
            with archive.open('MANIFEST.json') as f:
                return json.loads(str(f.read().decode()))
        except Exception as why:
            log.warning("Failed to load manifest data from json: %s", why)
            return None


def zip_import(path):
    if (not is_zip(path) or not os.path.isfile(path)):
        return None
    try:
        return zipimport.zipimporter(path)
    except zipimport.ZipImportError as why:
        log.error("ZIP import error: %s", why)
        return None


def _compatible_api_versions(api_versions):
    versions = [Version.from_string(v) for v in list(api_versions)]
    return set(versions) & set(picard.api_versions_tuple)


_plugin_dirs = []


def plugin_dirs():
    yield from _plugin_dirs


def init_default_plugin_dirs():
    # Add user plugin dir first
    if not os.path.exists(USER_PLUGIN_DIR):
        os.makedirs(USER_PLUGIN_DIR)
    register_plugin_dir(USER_PLUGIN_DIR)

    # Register system wide plugin dir
    if IS_FROZEN:
        toppath = sys.argv[0]
    else:
        toppath = os.path.abspath(__file__)

    topdir = os.path.dirname(toppath)
    plugin_dir = os.path.join(topdir, "plugins")
    register_plugin_dir(plugin_dir)


def register_plugin_dir(path):
    if path not in _plugin_dirs:
        _plugin_dirs.append(path)


def plugin_dir_for_path(path):
    for plugin_dir in plugin_dirs():
        try:
            if os.path.commonpath((path, plugin_dir)) == plugin_dir:
                return plugin_dir
        except ValueError:
            pass
    return path


class PluginManager(QtCore.QObject):

    plugin_installed = QtCore.pyqtSignal(PluginWrapper, bool)
    plugin_updated = QtCore.pyqtSignal(str, bool)
    plugin_removed = QtCore.pyqtSignal(str, bool)
    plugin_errored = QtCore.pyqtSignal(str, str, bool)
    updates_available = QtCore.pyqtSignal(list)

    def __init__(self, plugins_directory=None):
        super().__init__()
        self.plugins = []
        self._available_plugins = None  # None=never loaded, [] = empty
        if plugins_directory is None:
            plugins_directory = USER_PLUGIN_DIR
        self.plugins_directory = os.path.normpath(plugins_directory)
        init_default_plugin_dirs()

    @property
    def available_plugins(self):
        return self._available_plugins

    def plugin_error(self, name, error, *args, **kwargs):
        """Log a plugin loading error for the plugin `name` and signal the
        error via the `plugin_errored` signal.

        A string consisting of all `args` interpolated into `error` will be
        passed to the function given via the `log_func` keyword argument
        (default: log.error) and as the error message to the `plugin_errored`
        signal.

        Instead of using `args` the interpolation parameters can also be passed
        with the `params` keyword parameter. This is specifically useful to
        pass a dictionary when using named placeholders."""
        params = kwargs.get('params', args)
        if params:
            error = error % params
        log_func = kwargs.get('log_func', log.error)
        log_func(error)
        self.plugin_errored.emit(name, error, False)

    def _marked_for_update(self):
        for file in os.listdir(self.plugins_directory):
            if file.endswith(_UPDATE_SUFFIX):
                source_path = os.path.join(self.plugins_directory, file)
                target_path = strip_update_suffix(source_path)
                plugin_name = _plugin_name_from_path(target_path)
                if plugin_name:
                    yield (source_path, target_path, plugin_name)
                else:
                    log.error('Cannot get plugin name from %r', source_path)

    def handle_plugin_updates(self):
        for source_path, target_path, plugin_name in self._marked_for_update():
            self._remove_plugin(plugin_name)
            os.rename(source_path, target_path)
            log.debug("Updating plugin %r (%r))", plugin_name, target_path)

    def load_plugins_from_directory(self, plugindir):
        plugindir = os.path.normpath(plugindir)
        if not os.path.isdir(plugindir):
            log.info("Plugin directory %r doesn't exist", plugindir)
            return
        if plugindir == self.plugins_directory:
            # .update trick is only for plugins installed through the Picard UI
            # and only for plugins in plugins_directory (USER_PLUGIN_DIR by default)
            self.handle_plugin_updates()
        # now load found plugins
        names = set()
        for path in (os.path.join(plugindir, file) for file in os.listdir(plugindir)):
            name = _plugin_name_from_path(path)
            if name:
                names.add(name)
        log.debug("Looking for plugins in directory %r, %d names found",
                  plugindir,
                  len(names))
        for name in sorted(names):
            try:
                self._load_plugin(name)
            except Exception:
                self.plugin_error(name, _("Unable to load plugin '%s'"), name, log_func=log.exception)

    def _get_plugin_index_by_name(self, name):
        for index, plugin in enumerate(self.plugins):
            if name == plugin.module_name:
                return (plugin, index)
        return (None, None)

    def _load_plugin(self, name):
        existing_plugin, existing_plugin_index = self._get_plugin_index_by_name(name)
        if existing_plugin:
            log.debug("Ignoring already loaded plugin %r (version %r at %r)",
                existing_plugin.module_name,
                existing_plugin.version,
                existing_plugin.file)
            return

        spec = None
        module_pathname = None
        zip_importer = None
        manifest_data = None
        full_module_name = _PLUGIN_MODULE_PREFIX + name

        # Legacy loading of ZIP plugins. In Python >= 3.10 this is all handled
        # by PluginMetaPathFinder. Remove once Python 3.9 is no longer supported.
        if not hasattr(zipimport.zipimporter, 'find_spec'):
            (zip_importer, plugin_dir, module_pathname, manifest_data) = self._legacy_load_zip_plugin(name)

        if not module_pathname:
            spec = PluginMetaPathFinder().find_spec(full_module_name, [])
            if not spec or not spec.loader:
                errorfmt = _('Failed loading plugin "%(plugin)s"')
                self.plugin_error(name, errorfmt, params={
                    'plugin': name,
                })
                return None

            module_pathname = spec.origin
            if isinstance(spec.loader, zipimport.zipimporter):
                manifest_data = load_zip_manifest(spec.loader.archive)
            if os.path.basename(module_pathname) == '__init__.py':
                module_pathname = os.path.dirname(module_pathname)
            plugin_dir = plugin_dir_for_path(module_pathname)

        plugin = None
        try:
            if zip_importer:  # Legacy ZIP import for Python < 3.10
                plugin_module = zip_importer.load_module(full_module_name)
            else:
                plugin_module = importlib.util.module_from_spec(spec)
                # This is kind of a hack. The module will be in sys.modules
                # after exec_module has run. But if inside of the loaded plugin
                # there are relative imports it would load the same plugin
                # module twice. This executes the plugins code twice and leads
                # to potential side effects.
                sys.modules[full_module_name] = plugin_module
                try:
                    spec.loader.exec_module(plugin_module)
                except:  # noqa: E722
                    del sys.modules[full_module_name]
                    raise

            plugin = PluginWrapper(plugin_module, plugin_dir,
                                   file=module_pathname, manifest_data=manifest_data)
            compatible_versions = _compatible_api_versions(plugin.api_versions)
            if compatible_versions:
                log.debug("Loading plugin %r version %s, compatible with API: %s",
                          plugin.name,
                          plugin.version,
                          ", ".join([v.to_string(short=True) for v in
                                     sorted(compatible_versions)]))
                plugin.compatible = True
                setattr(picard.plugins, name, plugin_module)
                if existing_plugin:
                    self.plugins[existing_plugin_index] = plugin
                else:
                    self.plugins.append(plugin)
            else:
                errorfmt = _('Plugin "%(plugin)s" from "%(filename)s" is not '
                             'compatible with this version of Picard.')
                params = {'plugin': plugin.name, 'filename': plugin.file}
                self.plugin_error(plugin.name, errorfmt, params=params, log_func=log.warning)
        except VersionError as e:
            errorfmt = _('Plugin "%(plugin)s" has an invalid API version string: %(error)s')
            self.plugin_error(name, errorfmt, params={
                'plugin': name,
                'error': e,
            })
        except BaseException:
            errorfmt = _('Plugin "%(plugin)s"')
            self.plugin_error(name, errorfmt, log_func=log.exception,
                              params={'plugin': name})
        return plugin

    def _legacy_load_zip_plugin(self, name):
        for plugin_dir in plugin_dirs():
            zipfilename = os.path.join(plugin_dir, name + '.zip')
            zip_importer = zip_import(zipfilename)
            if zip_importer:
                if not zip_importer.find_module(name):
                    errorfmt = _('Failed loading zipped plugin "%(plugin)s" from "%(filename)s"')
                    self.plugin_error(name, errorfmt, params={
                        'plugin': name,
                        'filename': zipfilename,
                    })
                    return (None, None, None, None)
                module_pathname = zip_importer.get_filename(name)
                manifest_data = load_zip_manifest(zip_importer.archive)
                return (zip_importer, plugin_dir, module_pathname, manifest_data)
        return (None, None, None, None)

    def _get_existing_paths(self, plugin_name, fileexts):
        dirpath = os.path.join(self.plugins_directory, plugin_name)
        if not os.path.isdir(dirpath):
            dirpath = None
        filenames = {plugin_name + ext for ext in fileexts}
        filepaths = [os.path.join(self.plugins_directory, f)
                     for f in os.listdir(self.plugins_directory)
                     if f in filenames
                     ]
        return (dirpath, filepaths)

    def _remove_plugin_files(self, plugin_name, with_update=False):
        plugin_name = strip_zip_suffix(plugin_name)
        log.debug("Remove plugin files and dirs : %r", plugin_name)
        dirpath, filepaths = self._get_existing_paths(plugin_name, _FILEEXTS)
        if dirpath:
            if os.path.islink(dirpath):
                log.debug("Removing symlink %r", dirpath)
                os.remove(dirpath)
            elif os.path.isdir(dirpath):
                log.debug("Removing directory %r", dirpath)
                shutil.rmtree(dirpath)
        if filepaths:
            for filepath in filepaths:
                log.debug("Removing file %r", filepath)
                os.remove(filepath)
                if with_update:
                    update = filepath + _UPDATE_SUFFIX
                    if os.path.isfile(update):
                        log.debug("Removing file %r", update)
                        os.remove(update)

    def _remove_plugin(self, plugin_name, with_update=False):
        self._remove_plugin_files(plugin_name, with_update)
        _unregister_module_extensions(plugin_name)
        self.plugins = [p for p in self.plugins if p.module_name != plugin_name]

    def remove_plugin(self, plugin_name, with_update=False):
        self._remove_plugin(plugin_name, with_update=with_update)
        self.plugin_removed.emit(plugin_name, False)

    def _install_plugin_zip(self, plugin_name, plugin_data, update=False):
        # zipped module from download
        zip_plugin = plugin_name + '.zip'
        dst = os.path.join(self.plugins_directory, zip_plugin)
        if update:
            dst += _UPDATE_SUFFIX
            if os.path.isfile(dst):
                os.remove(dst)
        with tempfile.NamedTemporaryFile(dir=self.plugins_directory) as zipfile:
            zipfile.write(plugin_data)
            zipfile.flush()
            os.fsync(zipfile.fileno())
            try:
                os.link(zipfile.name, dst)
            except OSError:
                with open(dst, 'wb') as dstfile:
                    zipfile.seek(0)
                    shutil.copyfileobj(zipfile, dstfile)
            log.debug("Plugin (zipped) saved to %r", dst)

    def _install_plugin_file(self, path, update=False):
        dst = os.path.join(self.plugins_directory, os.path.basename(path))
        if update:
            dst += _UPDATE_SUFFIX
            if os.path.isfile(dst):
                os.remove(dst)
        shutil.copy2(path, dst)
        log.debug("Plugin (file) saved to %r", dst)

    def _install_plugin_dir(self, plugin_name, path, update=False):
        dst = os.path.join(self.plugins_directory, plugin_name)
        if update:
            dst += _UPDATE_SUFFIX
            if os.path.isdir(dst):
                shutil.rmtree(dst)
        shutil.copytree(path, dst)
        log.debug("Plugin (directory) saved to %r", dst)

    def install_plugin(self, path, update=False, plugin_name=None, plugin_data=None):
        """
            path is either:
                1) /some/dir/name.py
                2) /some/dir/name (directory containing __init__.py)
                3) /some/dir/name.zip (containing either 1 or 2)

        """
        assert path or plugin_name, "path is required if plugin_name is empty"

        if not plugin_name:
            plugin_name = _plugin_name_from_path(path)
        if plugin_name:
            try:
                if plugin_data:
                    self._install_plugin_zip(plugin_name, plugin_data, update=update)
                elif os.path.isfile(path):
                    self._install_plugin_file(path, update=update)
                elif os.path.isdir(path):
                    self._install_plugin_dir(plugin_name, path, update=update)
            except OSError as why:
                log.error("Unable to copy plugin '%s' to %r: %s", plugin_name, self.plugins_directory, why)
                return

            if not update:
                try:
                    installed_plugin = self._load_plugin(plugin_name)
                    if not installed_plugin:
                        raise RuntimeError("Failed loading newly installed plugin %s" % plugin_name)
                except Exception as e:
                    log.error("Unable to load plugin '%s': %s", plugin_name, e)
                    self._remove_plugin(plugin_name)
                else:
                    self.plugin_installed.emit(installed_plugin, False)
            else:
                self.plugin_updated.emit(plugin_name, False)

    def query_available_plugins(self, callback=None):
        self.tagger.webservice.get_url(
            url=PLUGINS_API['urls']['plugins'],
            handler=partial(self._plugins_json_loaded, callback=callback),
            priority=True,
            important=True,
        )

    def is_available(self, plugin_name):
        return any(p.module_name == plugin_name for p in self._available_plugins)

    def _plugins_json_loaded(self, response, reply, error, callback=None):
        if error:
            self.tagger.window.set_statusbar_message(
                N_("Error loading plugins list: %(error)s"),
                {'error': reply.errorString()},
                echo=log.error
            )
            self._available_plugins = []
        else:
            try:
                self._available_plugins = [PluginData(data, key) for key, data in
                                           response['plugins'].items()
                                           if _compatible_api_versions(data['api_versions'])]
            except (AttributeError, KeyError, TypeError):
                self._available_plugins = []
        if callback:
            callback()

    # pylint: disable=no-self-use
    def enabled(self, name):
        return True

    def _plugins_have_new_versions(self):
        """Compare available plugins versions with installed plugins ones
        and yield plugin names of plugins that have new versions"""
        if self.available_plugins is not None:
            available_versions = {p.module_name: p.version for p in self.available_plugins}
            for plugin in self.plugins:
                if plugin.module_name not in available_versions:
                    continue
                if available_versions[plugin.module_name] > plugin.version:
                    yield plugin.name

    def check_update(self):
        if self.available_plugins is None:
            self.query_available_plugins(self._notify_updates)
        else:
            self._notify_updates()

    def _notify_updates(self):
        plugins_with_updates = list(self._plugins_have_new_versions())
        if plugins_with_updates:
            self.updates_available.emit(plugins_with_updates)


class PluginMetaPathFinder(MetaPathFinder):
    def find_spec(self, fullname, path, target=None):
        if not fullname.startswith(_PLUGIN_MODULE_PREFIX):
            return None
        plugin_name = fullname[len(_PLUGIN_MODULE_PREFIX):]
        for plugin_dir in plugin_dirs():
            for file_path in self._plugin_file_paths(plugin_dir, plugin_name):
                if os.path.exists(file_path):
                    spec = self._spec_from_path(fullname, file_path)
                    if spec and spec.loader:
                        return spec

    def _spec_from_path(self, fullname, file_path):
        if file_path.endswith('.zip'):
            return self._spec_from_zip(fullname, file_path)
        else:
            return importlib.util.spec_from_file_location(fullname, file_path)

    def _spec_from_zip(self, fullname, file_path):
        zip_importer = zip_import(file_path)
        if zip_importer:
            return zip_importer.find_spec(fullname)

    @staticmethod
    def _plugin_file_paths(plugin_dir, plugin_name):
        for entry in _PACKAGE_ENTRIES:
            yield os.path.join(plugin_dir, plugin_name, entry)
        for ext in _FILEEXTS:
            # On Python < 3.10 ZIP file loading is handled in PluginManager._load_plugin
            if ext == '.zip' and not hasattr(zipimport.zipimporter, 'find_spec'):
                continue
            yield os.path.join(plugin_dir, plugin_name + ext)


sys.meta_path.append(PluginMetaPathFinder())



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2007 Lukáš Lalinský
# Copyright (C) 2014, 2018, 2020 Laurent Monin
# Copyright (C) 2017 Sambhav Kothari
# Copyright (C) 2018 Shen-Ta Hsieh
# Copyright (C) 2018-2020 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


# List of available user interface languages
UI_LANGUAGES = [
    # ('af', 'Afrikaans', N_('Afrikaans')),
    ('ar', 'العربية', N_('Arabic')),
    # ('ast', 'Asturian', N_('Asturian')),
    # ('bg', 'Български', N_('Bulgarian')),
    ('ca', 'Català', N_('Catalan')),
    ('cs', 'Čeština', N_('Czech')),
    # ('cy', 'Cymraeg', N_('Welsh')),
    ('da', 'Dansk', N_('Danish')),
    ('de', 'Deutsch', N_('German')),
    # ('de_CH', 'Deutsch (Schweiz)', N_('German (Switzerland)')),
    ('el', 'ελληνικά', N_('Greek')),
    ('en', 'English', N_('English')),
    ('en_AU', 'English (Australia)', N_('English (Australia)')),
    ('en_CA', 'English (Canada)', N_('English (Canada)')),
    ('en_GB', 'English (United Kingdom)', N_('English (United Kingdom)')),
    # ('eo', 'Esperanto', N_('Esperanto')),
    ('es', 'Español', N_('Spanish')),
    ('et', 'Eesti', N_('Estonian')),
    # ('fa', 'فارسی', N_('Persian')),
    ('fi', 'Suomi', N_('Finnish')),
    # ('fo', 'Føroyskt', N_('Faroese')),
    ('fr', 'Français', N_('French')),
    ('fr_CA', 'Français (Canada)', N_('French (Canada)')),
    # ('fy', 'Frysk', N_('Frisian')),
    ('gl', 'Galego', N_('Galician')),
    ('he', 'עברית', N_('Hebrew')),
    # ('hi', 'हिन्दी', N_('Hindi')),
    ('hu', 'Magyar', N_('Hungarian')),
    # ('id', 'Bahasa Indonesia', N_('Indonesian')),
    ('is', 'Íslenska', N_('Icelandic')),
    ('it', 'Italiano', N_('Italian')),
    ('ja', '日本語', N_('Japanese')),
    # ('kn', 'ಕನ್ನಡ', N_('Kannada')),
    ('ko', '한국어', N_('Korean')),
    ('lt', 'Lietuvių', N_('Lithuanian')),
    ('ms_MY', 'Bahasa Melayu (Malaysia)', N_('Malay (Malaysia)')),
    ('nb', 'Norsk bokmål', N_('Norwegian Bokmål')),
    # ('nds', 'Plattdüütsch', N_('Low German')),
    ('nl', 'Nederlands', N_('Dutch')),
    ('oc', 'Occitan', N_('Occitan')),
    ('pl', 'Polski', N_('Polish')),
    ('pt', 'Português', N_('Portuguese')),
    ('pt_BR', 'Português (Brasil)', N_('Portuguese (Brazil)')),
    ('ro', 'Română', N_('Romanian')),
    ('ru', 'Pyccĸий', N_('Russian')),
    # ('sco', 'Scots leid', N_('Scots')),
    ('sk', 'Slovenčina', N_('Slovak')),
    ('sl', 'Slovenščina', N_('Slovenian')),
    ('sq', 'Shqip', N_('Albanian')),
    # ('sr', 'Србин', N_('Serbian')),
    ('sv', 'Svenska', N_('Swedish')),
    # ('ta', 'தமிழ்', N_('Tamil')),
    ('tr', 'Türkçe', N_('Turkish')),
    ('uk', 'Украї́нська', N_('Ukrainian')),
    ('zh_CN', '中文（中国大陆）', N_('Chinese (China)')),
    ('zh_TW', '中文（台灣）', N_('Chinese (Taiwan)')),
]



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2013 Michael Wiencek
# Copyright (C) 2014 Lukáš Lalinský
# Copyright (C) 2014, 2017 Sophist-UK
# Copyright (C) 2014, 2017-2021 Laurent Monin
# Copyright (C) 2016-2017 Sambhav Kothari
# Copyright (C) 2019, 2021 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from functools import partial

from PyQt5 import QtCore

from picard import log
from picard.config import get_config


user_collections = {}


class Collection(QtCore.QObject):
    COLLECTION_ADD = 1
    COLLECTION_REMOVE = 2

    def __init__(self, collection_id, name, size):
        self.id = collection_id
        self.name = name
        self.pending = set()
        self.size = int(size)
        self.releases = set()
        mb_api = self.tagger.mb_api
        self.api_action = {
            self.COLLECTION_ADD: mb_api.put_to_collection,
            self.COLLECTION_REMOVE: mb_api.delete_from_collection,
        }

    def __repr__(self):
        return '<Collection %s (%s)>' % (self.name, self.id)

    def _modify(self, kind, ids, callback):
        ids -= self.pending
        if ids:
            self.pending |= ids
            when_done = partial(self._finished, kind, ids, callback)
            self.api_action[kind](self.id, list(ids), when_done)

    def add_releases(self, ids, callback):
        self._modify(self.COLLECTION_ADD, ids, callback)

    def remove_releases(self, ids, callback):
        self._modify(self.COLLECTION_REMOVE, ids, callback)

    def _finished(self, kind, ids, callback, document, reply, error):
        self.pending -= ids
        statusbar = self.tagger.window.set_statusbar_message
        if not error:
            count = len(ids)
            if kind == self.COLLECTION_ADD:
                self.releases |= ids
                self.size += count
                status_msg = ngettext(
                    'Added %(count)i release to collection "%(name)s"',
                    'Added %(count)i releases to collection "%(name)s"',
                    count)
                debug_msg = 'Added %(count)i releases to collection "%(name)s"'
            else:
                self.releases -= ids
                self.size -= count
                status_msg = ngettext(
                    'Removed %(count)i release from collection "%(name)s"',
                    'Removed %(count)i releases from collection "%(name)s"',
                    count)
                debug_msg = 'Removed %(count)i releases from collection "%(name)s"'
            callback()
            mparms = {'count': count, 'name': self.name}
            log.debug(debug_msg % mparms)
            statusbar(status_msg, mparms, translate=None, echo=None)
        else:
            statusbar(
                N_("Error while modifying collections: %(error)s"),
                {'error': reply.errorString()},
                echo=log.error
            )


def get_user_collection(collection_id, name, size, refresh=False):
    collection = user_collections.get(collection_id)
    if collection is None:
        collection = user_collections[collection_id] = Collection(collection_id, name, size)
    elif refresh:
        collection.name = name
        collection.size = size
    return collection


def load_user_collections(callback=None):
    tagger = QtCore.QObject.tagger

    def request_finished(document, reply, error):
        if error:
            tagger.window.set_statusbar_message(
                N_("Error loading collections: %(error)s"),
                {'error': reply.errorString()},
                echo=log.error
            )
            return
        if document and 'collections' in document:
            collection_list = document['collections']
            new_collections = set()

            for node in collection_list:
                if node['entity-type'] != 'release':
                    continue
                col_id = node['id']
                col_name = node['name']
                col_size = node['release-count']
                new_collections.add(col_id)
                get_user_collection(col_id, col_name, col_size, refresh=True)

            # remove collections which aren't returned by the web service anymore
            old_collections = set(user_collections) - new_collections
            for collection_id in old_collections:
                del user_collections[collection_id]

            log.debug("User collections: %r", [(k, v.name) for k, v in user_collections.items()])
        if callback:
            callback()

    if tagger.webservice.oauth_manager.is_authorized():
        tagger.mb_api.get_collection_list(partial(request_finished))
    else:
        user_collections.clear()


def add_release_to_user_collections(release_node):
    """Add album to collections"""
    # Check for empy collection list
    if 'collections' in release_node:
        release_id = release_node['id']
        config = get_config()
        username = config.persist['oauth_username'].lower()
        for node in release_node['collections']:
            if node['editor'].lower() == username:
                col_id = node['id']
                col_name = node['name']
                col_size = node['release-count']
                collection = get_user_collection(col_id, col_name, col_size)
                collection.releases.add(release_id)
                log.debug("Adding release %r to %r", release_id, collection)



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2017 Antonio Larrosa
# Copyright (C) 2017 Sambhav Kothari
# Copyright (C) 2018, 2020-2022 Philipp Wolfer
# Copyright (C) 2019-2021 Laurent Monin
# Copyright (C) 2021 Gabriel Ferreira
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from collections.abc import MutableSequence

from picard.config import get_config


class ImageList(MutableSequence):
    def __init__(self, iterable=()):
        self._images = list(iterable)
        self._hash_dict = {}
        self._changed = True

    def __len__(self):
        return len(self._images)

    def __getitem__(self, index):
        return self._images[index]

    def __setitem__(self, index, value):
        self._images[index] = value
        self._changed = True

    def __delitem__(self, index):
        del self._images[index]
        self._changed = True

    def insert(self, index, value):
        self._changed = True
        return self._images.insert(index, value)

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self._images)

    def _sorted(self):
        return sorted(self, key=lambda image: image.normalized_types())

    def __eq__(self, other):
        if len(self) != len(other):
            return False
        return self._sorted() == other._sorted()

    def copy(self):
        return self.__class__(self._images)

    def get_front_image(self):
        for img in self:
            if img.is_front_image():
                return img
        return None

    def to_be_saved_to_tags(self, settings=None):
        """Generator returning images to be saved to tags according to
           passed settings or config.setting
        """
        if settings is None:
            config = get_config()
            settings = config.setting
        if settings['save_images_to_tags']:
            only_one_front = settings['embed_only_one_front_image']
            for image in self:
                if not image.can_be_saved_to_tags:
                    continue
                if only_one_front:
                    if image.is_front_image():
                        yield image
                        break
                else:
                    yield image

    def strip_front_images(self):
        self._images = [image for image in self._images if not image.is_front_image()]
        self._changed = True

    def hash_dict(self):
        if self._changed:
            self._hash_dict = {img.datahash.hash(): img for img in self._images}
            self._changed = False
        return self._hash_dict


class ImageListState:
    def __init__(self):
        self.new_images = {}
        self.orig_images = {}
        self.sources = []
        self.has_common_new_images = True
        self.has_common_orig_images = True
        self.first_new_obj = True
        self.first_orig_obj = True
        # The next variables specify what will be updated
        self.update_new_metadata = False
        self.update_orig_metadata = False


def _process_images(state, src_obj, Track):
    # Check new images
    if state.update_new_metadata:
        src_dict = src_obj.metadata.images.hash_dict()
        prev_len = len(state.new_images)
        state.new_images.update(src_dict)
        if len(state.new_images) != prev_len:
            if not state.first_new_obj:
                state.has_common_new_images = False
        if state.first_new_obj:
            state.first_new_obj = False

    if state.update_orig_metadata and not isinstance(src_obj, Track):
        # Check orig images, but not for Tracks (which don't have a useful orig_metadata)
        src_dict = src_obj.orig_metadata.images.hash_dict()
        prev_len = len(state.orig_images)
        state.orig_images.update(src_dict)
        if len(state.orig_images) != prev_len:
            if not state.first_orig_obj:
                state.has_common_orig_images = False
        if state.first_orig_obj:
            state.first_orig_obj = False


def _update_state(obj, state):
    from picard.track import Track

    changed = False
    for src_obj in state.sources:
        _process_images(state, src_obj, Track)

    if state.update_new_metadata:
        updated_images = ImageList(state.new_images.values())
        changed |= updated_images.hash_dict().keys() != obj.metadata.images.hash_dict().keys()
        obj.metadata.images = updated_images
        obj.metadata.has_common_images = state.has_common_new_images

    if state.update_orig_metadata:
        updated_images = ImageList(state.orig_images.values())
        changed |= updated_images.hash_dict().keys() != obj.orig_metadata.images.hash_dict().keys()
        obj.orig_metadata.images = updated_images
        obj.orig_metadata.has_common_images = state.has_common_orig_images

    return changed


# TODO: use functools.singledispatch when py3 is supported
def _get_state(obj):
    from picard.album import Album

    from picard.ui.item import FileListItem

    state = ImageListState()

    if isinstance(obj, Album):
        for track in obj.tracks:
            state.sources.append(track)
            state.sources += track.files
        state.sources += obj.unmatched_files.files
        state.update_new_metadata = True
        state.update_orig_metadata = True
    elif isinstance(obj, FileListItem):
        state.sources = obj.files
        state.update_new_metadata = True
        state.update_orig_metadata = True

    return state


def _get_metadata_images(state, sources):
    new_images = set()
    orig_images = set()
    for s in sources:
        if state.update_new_metadata:
            new_images = new_images.union(s.metadata.images)
        if state.update_orig_metadata:
            orig_images = orig_images.union(s.orig_metadata.images)
    return (new_images, orig_images)


def update_metadata_images(obj):
    """Update the metadata images `obj` based on its children.

    Based on the type of `obj` this will update `obj.metadata.images` to
    represent the metadata images of all children (`Track` or `File` objects).

    This method will iterate over all children and completely rebuild
    `obj.metadata.images`. Whenever possible the more specific functions
    `add_metadata_images` or `remove_metadata_images` should be used.

    Args:
        obj: A `Cluster`, `Album` or `Track` object with `metadata` property
    Returns:
        bool: True, if images where changed, False otherwise
    """
    return _update_state(obj, _get_state(obj))


def _add_images(metadata, added_images):
    if not added_images:
        return False

    current_images = set(metadata.images)
    if added_images.isdisjoint(current_images):
        metadata.images = ImageList(current_images.union(added_images))
        metadata.has_common_images = False
        return True

    return False


def add_metadata_images(obj, added_sources):
    """Add the images in the metadata of `added_sources` to the metadata of `obj`.

    Args:
        obj: A `Cluster`, `Album` or `Track` object with `metadata` property
        added_sources: List of child objects (`Track` or `File`) which's metadata images should be added to `obj`
    """
    state = _get_state(obj)
    (added_new_images, added_orig_images) = _get_metadata_images(state, added_sources)
    changed = False

    if state.update_new_metadata:
        changed |= _add_images(obj.metadata, added_new_images)
    if state.update_orig_metadata:
        changed |= _add_images(obj.orig_metadata, added_orig_images)

    return changed


def _remove_images(metadata, sources, removed_images):
    """Removes `removed_images` from metadata `images`, but only if they are not included in `sources`.

    Args:
        metadata: `Metadata` object from which images should be removed
        sources: List of source `Metadata` objects
        removed_images: Set of `CoverArt` proposed for removal from `metadata`
    """
    if not metadata.images or not removed_images:
        return

    if not sources:
        metadata.images = ImageList()
        metadata.has_common_images = True
        return

    current_images = set(metadata.images)

    if metadata.has_common_images and current_images == removed_images:
        return

    common_images = True  # True, if all children share the same images
    previous_images = None

    # Iterate over all sources and check whether the images proposed to be
    # removed are used in any sources. Images used in existing sources
    # must not be removed.
    for source_metadata in sources:
        source_images = set(source_metadata.images)
        if previous_images and common_images and previous_images != source_images:
            common_images = False
        previous_images = set(source_metadata.images)  # Remember for next iteration
        removed_images = removed_images.difference(source_images)
        if not removed_images and not common_images:
            return  # No images left to remove, abort immediately

    metadata.images = ImageList(current_images.difference(removed_images))
    metadata.has_common_images = common_images


def remove_metadata_images(obj, removed_sources):
    """Remove the images in the metadata of `removed_sources` from the metadata of `obj`.

    Args:
        obj: A `Cluster`, `Album` or `Track` object with `metadata` property
        removed_sources: List of child objects (`Track` or `File`) which's metadata images should be removed from `obj`
    """
    from picard.track import Track

    state = _get_state(obj)
    (removed_new_images, removed_orig_images) = _get_metadata_images(state, removed_sources)

    if state.update_new_metadata:
        sources = [s.metadata for s in state.sources]
        _remove_images(obj.metadata, sources, removed_new_images)
    if state.update_orig_metadata:
        sources = [s.orig_metadata for s in state.sources if not isinstance(s, Track)]
        _remove_images(obj.orig_metadata, sources, removed_orig_images)



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2006-2007, 2009 Lukáš Lalinský
# Copyright (C) 2014 m42i
# Copyright (C) 2020-2022 Laurent Monin
# Copyright (C) 2020-2022 Philipp Wolfer
# Copyright (C) 2021-2022 Bob Swift
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


import re
import unicodedata

from PyQt5 import (
    QtCore,
    QtGui,
)
from PyQt5.QtCore import Qt
from PyQt5.QtGui import (
    QCursor,
    QKeySequence,
    QTextCursor,
)
from PyQt5.QtWidgets import (
    QAction,
    QCompleter,
    QTextEdit,
    QToolTip,
)

from picard.config import (
    BoolOption,
    get_config,
)
from picard.const.sys import IS_MACOS
from picard.script import (
    ScriptFunctionDocError,
    script_function_documentation,
    script_function_names,
)
from picard.util.tags import (
    PRESERVED_TAGS,
    TAG_NAMES,
    display_tag_name,
)

from picard.ui import FONT_FAMILY_MONOSPACE
from picard.ui.theme import theme


EXTRA_VARIABLES = (
    '~absolutetracknumber',
    '~albumartists_sort',
    '~albumartists',
    '~artists_sort',
    '~datatrack',
    '~discpregap',
    '~multiartist',
    '~musicbrainz_discids',
    '~musicbrainz_tracknumber',
    '~performance_attributes',
    '~pregap',
    '~primaryreleasetype',
    '~rating',
    '~recording_firstreleasedate',
    '~recordingcomment',
    '~recordingtitle',
    '~releasecomment',
    '~releasecountries',
    '~releasegroup_firstreleasedate',
    '~releasegroup',
    '~releasegroupcomment',
    '~releaselanguage',
    '~secondaryreleasetype',
    '~silence',
    '~totalalbumtracks',
    '~video',
)


def find_regex_index(regex, text, start=0):
    match = regex.search(text[start:])
    return start + match.start() if match else -1


class TaggerScriptSyntaxHighlighter(QtGui.QSyntaxHighlighter):

    def __init__(self, document):
        super().__init__(document)
        syntax_theme = theme.syntax_theme
        self.func_re = re.compile(r"\$(?!noop)[_a-zA-Z0-9]*\(")
        self.func_fmt = QtGui.QTextCharFormat()
        self.func_fmt.setFontWeight(QtGui.QFont.Weight.Bold)
        self.func_fmt.setForeground(syntax_theme.func)
        self.var_re = re.compile(r"%[_a-zA-Z0-9:]*%")
        self.var_fmt = QtGui.QTextCharFormat()
        self.var_fmt.setForeground(syntax_theme.var)
        self.unicode_re = re.compile(r"\\u[a-fA-F0-9]{4}")
        self.unicode_fmt = QtGui.QTextCharFormat()
        self.unicode_fmt.setForeground(syntax_theme.escape)
        self.escape_re = re.compile(r"\\[^u]")
        self.escape_fmt = QtGui.QTextCharFormat()
        self.escape_fmt.setForeground(syntax_theme.escape)
        self.special_re = re.compile(r"[^\\][(),]")
        self.special_fmt = QtGui.QTextCharFormat()
        self.special_fmt.setForeground(syntax_theme.special)
        self.bracket_re = re.compile(r"[()]")
        self.noop_re = re.compile(r"\$noop\(")
        self.noop_fmt = QtGui.QTextCharFormat()
        self.noop_fmt.setFontWeight(QtGui.QFont.Weight.Bold)
        self.noop_fmt.setFontItalic(True)
        self.noop_fmt.setForeground(syntax_theme.noop)
        self.rules = [
            (self.func_re, self.func_fmt, 0, -1),
            (self.var_re, self.var_fmt, 0, 0),
            (self.unicode_re, self.unicode_fmt, 0, 0),
            (self.escape_re, self.escape_fmt, 0, 0),
            (self.special_re, self.special_fmt, 1, -1),
        ]

    def highlightBlock(self, text):
        self.setCurrentBlockState(0)

        for expr, fmt, a, b in self.rules:
            for match in expr.finditer(text):
                index = match.start()
                length = match.end() - match.start()
                self.setFormat(index + a, length + b, fmt)

        # Ignore everything if we're already in a noop function
        index = find_regex_index(self.noop_re, text) if self.previousBlockState() <= 0 else 0
        open_brackets = self.previousBlockState() if self.previousBlockState() > 0 else 0
        text_length = len(text)
        while index >= 0:
            next_index = find_regex_index(self.bracket_re, text, index)

            # Skip escaped brackets
            if next_index > 0 and text[next_index - 1] == '\\':
                next_index += 1

            # Reached end of text?
            if next_index >= text_length:
                self.setFormat(index, text_length - index, self.noop_fmt)
                break

            if next_index > -1 and text[next_index] == '(':
                open_brackets += 1
            elif next_index > -1 and text[next_index] == ')':
                open_brackets -= 1

            if next_index > -1:
                self.setFormat(index, next_index - index + 1, self.noop_fmt)
            elif next_index == -1 and open_brackets > 0:
                self.setFormat(index, text_length - index, self.noop_fmt)

            # Check for next noop operation, necessary for multiple noops in one line
            if open_brackets == 0:
                next_index = find_regex_index(self.noop_re, text, next_index)

            index = next_index + 1 if next_index > -1 and next_index < text_length else -1

        self.setCurrentBlockState(open_brackets)


class ScriptCompleter(QCompleter):
    def __init__(self, parent=None):
        super().__init__(sorted(self.choices), parent)
        self.setCompletionMode(QCompleter.CompletionMode.UnfilteredPopupCompletion)
        self.highlighted.connect(self.set_highlighted)
        self.last_selected = ''

    @property
    def choices(self):
        yield from {'$' + name for name in script_function_names()}
        yield from {'%' + name.replace('~', '_') + '%' for name in self.all_tags}

    @property
    def all_tags(self):
        yield from TAG_NAMES.keys()
        yield from PRESERVED_TAGS
        yield from EXTRA_VARIABLES

    def set_highlighted(self, text):
        self.last_selected = text

    def get_selected(self):
        return self.last_selected


class DocumentedScriptToken:

    allowed_chars = re.compile('[A-Za-z0-9_]')

    def __init__(self, doc, cursor_position):
        self._doc = doc
        self._cursor_position = cursor_position

    def is_start_char(self, char):
        return False

    def is_allowed_char(self, char, position):
        return self.allowed_chars.match(char)

    def get_tooltip(self, position):
        return None

    def _read_text(self, position, count):
        text = ''
        while count:
            char = self._doc.characterAt(position)
            if not char:
                break
            text += char
            count -= 1
            position += 1
        return text

    def _read_allowed_chars(self, position):
        doc = self._doc
        text = ''
        while True:
            char = doc.characterAt(position)
            if not self.allowed_chars.match(char):
                break
            text += char
            position += 1
        return text


class FunctionScriptToken(DocumentedScriptToken):

    def is_start_char(self, char):
        return char == '$'

    def get_tooltip(self, position):
        if self._doc.characterAt(position) != '$':
            return None
        function = self._read_allowed_chars(position + 1)
        try:
            return script_function_documentation(function, 'html')
        except ScriptFunctionDocError:
            return None


class VariableScriptToken(DocumentedScriptToken):

    allowed_chars = re.compile('[A-Za-z0-9_:]')

    def is_start_char(self, char):
        return char == '%'

    def get_tooltip(self, position):
        if self._doc.characterAt(position) != '%':
            return None
        tag = self._read_allowed_chars(position + 1)
        return display_tag_name(tag)


class UnicodeEscapeScriptToken(DocumentedScriptToken):

    allowed_chars = re.compile('[uA-Fa-f0-9]')
    unicode_escape_sequence = re.compile('^\\\\u[a-fA-F0-9]{4}$')

    def is_start_char(self, char):
        return char == '\\'

    def is_allowed_char(self, char, position):
        return self.allowed_chars.match(char) and self._cursor_position - position < 6

    def get_tooltip(self, position):
        text = self._read_text(position, 6)
        if self.unicode_escape_sequence.match(text):
            codepoint = int(text[2:], 16)
            char = chr(codepoint)
            try:
                tooltip = unicodedata.name(char)
            except ValueError:
                tooltip = f'U+{text[2:].upper()}'
            if unicodedata.category(char)[0] != "C":
                tooltip += f': "{char}"'
            return tooltip
        return None


def _clean_text(text):
    return "".join(_replace_control_chars(text))


def _replace_control_chars(text):
    simple_ctrl_chars = {'\n', '\r', '\t'}
    for ch in text:
        if ch not in simple_ctrl_chars and unicodedata.category(ch)[0] == "C":
            yield '\\u' + hex(ord(ch))[2:].rjust(4, '0')
        else:
            yield ch


class ScriptTextEdit(QTextEdit):
    autocomplete_trigger_chars = re.compile('[$%A-Za-z0-9_]')

    options = [
        BoolOption('persist', 'script_editor_wordwrap', False),
        BoolOption('persist', 'script_editor_tooltips', True),
    ]

    def __init__(self, parent):
        super().__init__(parent)
        config = get_config()
        self.highlighter = TaggerScriptSyntaxHighlighter(self.document())
        self.enable_completer()
        self.setFontFamily(FONT_FAMILY_MONOSPACE)
        self.setMouseTracking(True)
        self.setAcceptRichText(False)
        self.wordwrap_action = QAction(_("&Word wrap script"), self)
        self.wordwrap_action.setToolTip(_("Word wrap long lines in the editor"))
        self.wordwrap_action.triggered.connect(self.update_wordwrap)
        self.wordwrap_action.setShortcut(QKeySequence(_("Ctrl+Shift+W")))
        self.wordwrap_action.setCheckable(True)
        self.wordwrap_action.setChecked(config.persist['script_editor_wordwrap'])
        self.update_wordwrap()
        self.addAction(self.wordwrap_action)
        self._show_tooltips = config.persist['script_editor_tooltips']
        self.show_tooltips_action = QAction(_("Show help &tooltips"), self)
        self.show_tooltips_action.setToolTip(_("Show tooltips for script elements"))
        self.show_tooltips_action.triggered.connect(self.update_show_tooltips)
        self.show_tooltips_action.setShortcut(QKeySequence(_("Ctrl+Shift+T")))
        self.show_tooltips_action.setCheckable(True)
        self.show_tooltips_action.setChecked(self._show_tooltips)
        self.addAction(self.show_tooltips_action)
        self.textChanged.connect(self.update_tooltip)

    def contextMenuEvent(self, event):
        menu = self.createStandardContextMenu()
        menu.addSeparator()
        menu.addAction(self.wordwrap_action)
        menu.addAction(self.show_tooltips_action)
        menu.exec_(event.globalPos())

    def mouseMoveEvent(self, event):
        if self._show_tooltips:
            tooltip = self.get_tooltip_at_mouse_position(event.pos())
            if not tooltip:
                QToolTip.hideText()
            self.setToolTip(tooltip)
        return super().mouseMoveEvent(event)

    def update_tooltip(self):
        if self.underMouse() and self.toolTip():
            position = self.mapFromGlobal(QCursor.pos())
            tooltip = self.get_tooltip_at_mouse_position(position)
            if tooltip != self.toolTip():
                # Hide tooltip if the entity causing this tooltip
                # was moved away from the mouse position
                QToolTip.hideText()
                self.setToolTip(tooltip)

    def get_tooltip_at_mouse_position(self, position):
        cursor = self.cursorForPosition(position)
        return self.get_tooltip_at_cursor(cursor)

    def get_tooltip_at_cursor(self, cursor):
        position = cursor.position()
        doc = self.document()
        documented_tokens = {
            FunctionScriptToken(doc, position),
            VariableScriptToken(doc, position),
            UnicodeEscapeScriptToken(doc, position)
        }
        while position >= 0 and documented_tokens:
            char = doc.characterAt(position)
            for token in list(documented_tokens):
                if token.is_start_char(char):
                    return token.get_tooltip(position)
                elif not token.is_allowed_char(char, position):
                    documented_tokens.remove(token)
            position -= 1
        return None

    def insertFromMimeData(self, source):
        text = _clean_text(source.text())
        # Create a new data object, as modifying the existing one does not
        # work on Windows if copying from outside the Qt app.
        source = QtCore.QMimeData()
        source.setText(text)
        return super().insertFromMimeData(source)

    def setPlainText(self, text):
        super().setPlainText(text)
        self.update_wordwrap()

    def update_wordwrap(self):
        """Toggles wordwrap in the script editor
        """
        wordwrap = self.wordwrap_action.isChecked()
        config = get_config()
        config.persist['script_editor_wordwrap'] = wordwrap
        if wordwrap:
            self.setLineWrapMode(QTextEdit.LineWrapMode.WidgetWidth)
        else:
            self.setLineWrapMode(QTextEdit.LineWrapMode.NoWrap)

    def update_show_tooltips(self):
        """Toggles wordwrap in the script editor
        """
        self._show_tooltips = self.show_tooltips_action.isChecked()
        config = get_config()
        config.persist['script_editor_tooltips'] = self._show_tooltips
        if not self._show_tooltips:
            QToolTip.hideText()
            self.setToolTip('')

    def enable_completer(self):
        self.completer = ScriptCompleter()
        self.completer.setWidget(self)
        self.completer.activated.connect(self.insert_completion)
        self.popup_shown = False

    def insert_completion(self, completion):
        if not completion:
            return
        tc = self.cursor_select_word()
        if completion.startswith('$'):
            completion += '('
        tc.insertText(completion)
        # Peek at the next character to include it in the replacement
        if not tc.atEnd():
            pos = tc.position()
            tc = self.textCursor()
            tc.setPosition(pos + 1, QTextCursor.MoveMode.KeepAnchor)
            first_char = completion[0]
            next_char = tc.selectedText()
            if (first_char == '$' and next_char == '(') or (first_char == '%' and next_char == '%'):
                tc.removeSelectedText()
            else:
                tc.setPosition(pos)  # Reset position
        self.setTextCursor(tc)
        self.popup_hide()

    def popup_hide(self):
        self.completer.popup().hide()

    def cursor_select_word(self, full_word=True):
        tc = self.textCursor()
        current_position = tc.position()
        tc.select(QTextCursor.SelectionType.WordUnderCursor)
        selected_text = tc.selectedText()
        # Check for start of function or end of variable
        if current_position > 0 and selected_text and selected_text[0] in {'(', '%'}:
            current_position -= 1
            tc.setPosition(current_position)
            tc.select(QTextCursor.SelectionType.WordUnderCursor)
            selected_text = tc.selectedText()
        start = tc.selectionStart()
        end = tc.selectionEnd()
        if current_position < start or current_position > end:
            # If the cursor is between words WordUnderCursor will select the
            # previous word. Reset the selection if the new selection is
            # outside the old cursor position.
            tc.setPosition(current_position)
            selected_text = tc.selectedText()
        if not selected_text.startswith('$') and not selected_text.startswith('%'):
            # Update selection to include the character before the
            # selected word to include the $ or %.
            tc.setPosition(start - 1 if start > 0 else 0)
            tc.setPosition(end, QTextCursor.MoveMode.KeepAnchor)
            selected_text = tc.selectedText()
            # No match, reset position (otherwise we could replace an additional character)
            if not selected_text.startswith('$') and not selected_text.startswith('%'):
                tc.setPosition(start)
                tc.setPosition(end, QTextCursor.MoveMode.KeepAnchor)
        if not full_word:
            tc.setPosition(current_position, QTextCursor.MoveMode.KeepAnchor)
        return tc

    def keyPressEvent(self, event):
        if self.completer.popup().isVisible():
            if event.key() in {Qt.Key.Key_Tab, Qt.Key.Key_Return, Qt.Key.Key_Enter}:
                self.completer.activated.emit(self.completer.get_selected())
                return

        super().keyPressEvent(event)
        self.handle_autocomplete(event)

    def handle_autocomplete(self, event):
        # Only trigger autocomplete on actual text input or if the user explicitly
        # requested auto completion with Ctrl+Space (Control+Space on macOS)
        modifier = QtCore.Qt.KeyboardModifier.MetaModifier if IS_MACOS else QtCore.Qt.KeyboardModifier.ControlModifier
        force_completion_popup = event.key() == QtCore.Qt.Key.Key_Space and event.modifiers() & modifier
        if not (force_completion_popup
                or event.key() in {Qt.Key.Key_Backspace, Qt.Key.Key_Delete}
                or self.autocomplete_trigger_chars.match(event.text())):
            self.popup_hide()
            return

        tc = self.cursor_select_word(full_word=False)
        selected_text = tc.selectedText()
        if force_completion_popup or (selected_text and selected_text[0] in {'$', '%'}):
            self.completer.setCompletionPrefix(selected_text)
            popup = self.completer.popup()
            popup.setCurrentIndex(self.completer.currentIndex())

            cr = self.cursorRect()
            cr.setWidth(
                popup.sizeHintForColumn(0)
                + popup.verticalScrollBar().sizeHint().width()
            )
            self.completer.complete(cr)
        else:
            self.popup_hide()
